// SPDX-License-Identifier: LGPL-2.1-or-later
/*
 * This file is part of libnvme.
 * Copyright (c) 2021 Code Construct Pty Ltd
 *
 * Authors: Jeremy Kerr <jk@codeconstruct.com.au>
 */

#ifndef _LIBNVME_MI_H
#define _LIBNVME_MI_H

#ifdef __cplusplus
extern "C" {
#endif

#include "nvme/types.h"
#include "nvme/mi.h"
#include "nvme/log.h"

#ifdef __cplusplus
}
#endif

#endif /* _LIBNVME_MI_H */
